﻿var Promise = require('promise');
var path = require('path');
var fs = require('fs');
var spawn = require('child_process').spawn;
var diffDependencies = require('./dependencies_diff/diff-dependencies.js').diffDependencies;

function BuildScannerProcess(cfg, diffService, technologies, sourceConrolProviders, parsedDependencies) {
    
    function createBuildDiffPromiseChain(buildArguments) {
        var handlerParams = {
            buildArguments: buildArguments,
            cfg: cfg,
            finalOutput: {
                "meta": {
                    "generated": new Date().valueOf(),
                    "appName": buildArguments.appname,
                    "build": buildArguments.build,
                    "branch": buildArguments.branch,
                    
                    "commit": buildArguments.commit,
                    "customerId": cfg.customerId,                    
                    "technology": buildArguments.technology[0] || "Java"
                },
                brokenBuild: false, //will be set to true on parse errors, etc.
                files: [],
                dependencies: parsedDependencies || []
            },
            previousMapping: null
        };
        
        var scmInstance;
        
        if (buildArguments.workspacepath) {
            var scmType = sourceConrolProviders[buildArguments.scm];
            if (scmType) {
                handlerParams.finalOutput.meta.scm = buildArguments.scm;
                scmInstance = new scmType(cfg, buildArguments.workspacePath);
            }
        }
        
        if (!buildArguments.commit && scmInstance && !scmInstance.detectCommitVersion) {
            return Promise.reject("'commit' command-line argument is required, since automatic detection by the " + buildArguments.scm + " SCM is not supported");
        }
        
        var promiseChain = (new Promise(function (resolve) { resolve(null); }));
        if (buildArguments.workspacepath) {
            
            //If SCM is used, load branch history & get current version
            if (scmInstance) {
                promiseChain = promiseChain.
                then(function () {
                    return scmInstance.getBranchHistory(handlerParams).then(function (branchLog) {
                        handlerParams.finalOutput.meta.history = branchLog;
                    });
                }).
                then(function () {
                    return scmInstance.detectCommitVersion(handlerParams).then(function (commit) {
                        handlerParams.finalOutput.meta.commit = commit;
                    });
                });
            };
            
            //Java Only
            if (buildArguments.technology) {
                promiseChain = promiseChain.then(function () {
                    return javaHandler(handlerParams);
                });
            }
        };
        
        //4. After going over all technology handlers, submit the final output to the server
        promiseChain = promiseChain.then(function () {
            if (handlerParams.finalOutput.brokenBuild) {
                return Promise.reject("Build agent encountered critical errors that prevent this build from being submitted.");
            }
            else {
                delete handlerParams.finalOutput.brokenBuild;
                return diffService.submitMapping(handlerParams.finalOutput);
            }
        });
        
        promiseChain = promiseChain.then(function () {
            console.log("Build agent finished");
            return Promise.resolve();
        }, function (err) {
            console.log("Build agent error: " + err);
            return Promise.resolve();
        });
        
        return promiseChain;
    }
    
    this.run = function run(buildArguments) {
        if (buildArguments.technology.length != 1 || buildArguments.technology[0].toLowerCase() != "java")
            return Promise.reject("Only java is supported");
        
        return createBuildDiffPromiseChain(buildArguments);
    }
        
    //JAVA
    
    function javaHandler(handlerParams) {
        var opts = handlerParams.buildArguments;
        var previousMapping = handlerParams.previousMapping;
        var finalOutput = handlerParams.finalOutput;
        var diffData = handlerParams.diffData;
        
        var p = new Promise(function (resolve, reject) {
            handlerParams.cfg.javaHome = handlerParams.cfg.javaHome || process.env.JAVA_HOME;
            handlerParams.cfg.javaClassPrefix = handlerParams.cfg.javaClassPrefix || '';
            if (!handlerParams.cfg.jar) {
                console.log('jar file is missing in sealights.json. Use sl-config-cia to specify it.');
                reject(null);
                return;
            }
            var executable = path.join(handlerParams.cfg.javaHome, 'bin/java');
            console.log('Java home = ' + handlerParams.cfg.javaHome);
            console.log('JAR location = ' + handlerParams.cfg.jar);
            console.log('Java prefix = ' + handlerParams.cfg.javaClassPrefix);
            console.log('Java executable = ' + executable);
            
            var tempFilename = 'mapping-' + new Date().valueOf() + '.json';
            //find jars in path opts.workspacepath
            console.log('Searching for *.jar/*.war files in workspace path "' + opts.workspacepath + '". Recursive search: ' + opts.recursive);
            
            var options = {
                isRecursive: opts.recursive, 
                filter: isJarOrWar
            };
            
            searchFiles(opts.workspacepath, onSearchComplete , options);
            
            function onSearchComplete(err, files) {
                if (err) {
                    console.log('Could not find *.jar/*.war files due to error:' + err);
                    reject(err);
                    return;
                }
                
                var jars = files;
                if (jars.length == 0) {
                    console.log('Could not find any *.jar/*.war files in workspace, aborting');
                    reject(err);
                    return;
                }
                
                jars.forEach(function (fullFilename) {
                    console.log('Found *.jar/*.war: ' + fullFilename)
                });
                
                try {
                    var childProcess = spawn(executable, ['-jar', handlerParams.cfg.jar, tempFilename].concat(jars), { stdio: 'inherit' });
                    //console.log('Spawned java pid=' + childProcess.pid);
                    childProcess.on('exit', function (code, signal) {
                        console.log('Java mapper existed with code=' + code + ', signal=' + signal);
                        if (code || signal) {
                            console.log('ERROR.');
                            reject({ code: code, signal: signal });
                            return;
                        }
                        try {
                            fs.readFile(tempFilename, 'utf8', function (err, contents) {
                                if (err) {
                                    console.log('Failed reading temp file ' + tempFilename);
                                    console.log('Error: ' + err);
                                    reject(err);
                                    return;
                                }
                                
                                try {
                                    handlerParams.finalOutput.files = JSON.parse(contents).files;
                                }
                            catch (e) {
                                    console.log('Failed parsing temp file ' + tempFilename);
                                    console.log('Error: ' + e);
                                    console.log('contents: ' + contents);
                                    reject(e);
                                    return;
                                }
                                
                                try {
                                    var fileFilter = createFileFilter(handlerParams.cfg.javaClassPrefix);
                                    handlerParams.finalOutput.files = handlerParams.finalOutput.files.filter(function (f) {
                                        return fileFilter(f.filename);
                                    });
                                    resolve(true);
                                }
                            catch (e) {
                                    console.log('Failed diffing between current and previous builds: ' + e);
                                    reject(e);
                                    return;
                                }

                            });
                        } catch (e) {
                            console.log('Could not read temp file ' + tempFilename + ": " + e);
                            reject(e);
                            return;
                        }
                    });
                } catch (e) {
                    console.log('Failed launching java mapper:' + e);
                    reject(e);
                    return;
                }
            }
        });
        
        return p;
    }
    
    function createFileFilter(prefix) {
        prefix = prefix || '';
        if (!prefix.length) return function () { return true; }; //no filter - everything goes
        var prefixParts = prefix.split(',');
        var partCount = prefixParts.length;
        return function (filename) {
            try {
                //if (partCount == 0) return true; //no filter - everything goes
                for (var i = 0; i < partCount; i++) {
                    if (filename.indexOf(prefixParts[i]) === 0) {
                        return true;
                    }
                }
                return false;
            } catch (e) {
                console.log(e);
                return false;
            }
        };
    }
    
    function isJarOrWar(filename) {
        var fileExtension = path.extname(filename).toLowerCase();
        return (fileExtension == ".jar" || fileExtension == ".war");
    }
    
    function searchFiles(dir, doneCallback, options) {
        var filterFunction = options.filter;
        var results = [];
        //console.log("[search files] Dir:" + dir);
        fs.readdir(dir, function (err, filesOrFolders) {
            //console.log("[search files] error:" + err + ", filesOrFolders:" + filesOrFolders);
            if (err) {
                return doneCallback(err);
            }
            
            var i = 0;
            
            (function processNextFileOrFolder() {
                var fileOrFolder = filesOrFolders[i++];
                if (!fileOrFolder) {
                    //No more files / folders. Return the results.
                    return doneCallback(null, results);
                }
                
                //Create a full path.
                fileOrFolder = path.join(dir, fileOrFolder);
                
                fs.stat(fileOrFolder, function (err, stat) {
                    //If that's a directory, continue the search.
                    if (options.isRecursive && stat && stat.isDirectory()) {
                        searchFiles(fileOrFolder, function (err, res) {
                            results = results.concat(res);
                            processNextFileOrFolder();
                        }, options);
                    } else {
                        //The current item is a file. Add it to the result.
                        if (!filterFunction)
                            results.push(fileOrFolder);
                        else if (filterFunction(fileOrFolder))
                            results.push(fileOrFolder);
                        
                        processNextFileOrFolder();
                    }
                });
            })();
        });
    };

}

module.exports.BuildScannerProcess = BuildScannerProcess;