﻿var nopt = require("nopt"),
    noptTypes = require("./nopt-types.js"),
    noptUsage = require("nopt-usage"),
    path = require("path"),
    DiffService = require("./diff-service.js").DiffService,
    NewDiffService = require("./new-diff-service.js").NewDiffService,
    Promise = require('promise'),
    parseDependencies = require('./dependency-parser.js').parseDependencies,
    BuildDiffProcess = require("./build-diff-process.js").BuildDiffProcess,
    BuildScannerProcess = require("./build-scanner-process.js").BuildScannerProcess,
    cfgLoader = require('./configuration-loader.js'),
    needle = require('needle'),
    readJsonSync = require('read-json-sync');


function CIA(technologies, sourceConrolProviders) {
    this.getTechnologies = function () { return technologies; };
    this.getSourceConrolProviders = function () { return sourceConrolProviders; };
}

CIA.prototype.run = function run(callback) {
    console.log('SeaLights Build agent ' + require("../package.json").version);
    nopt.typeDefs.technology = noptTypes.technology;
    nopt.typeDefs.git = noptTypes.git;
    
    var hasMissingArguments = false;
    nopt.invalidHandler = function (key, val, types) {
        hasMissingArguments = true;
        console.log('Invalid value specified for ' + key + ': ' + val);
    };
    
    var knownOpts = {
        "workspacepath" : path
        , "outputpath": path
        , "branch" : String
        , "build" : String
        , "commit" : String
        , "appname" : String
        , "config" : String
        , "newBuildScanner" : Boolean
        , "technology": ["technology", Array]
        , "scm" : "scm"
        , "help": Boolean
        , "dependency": [String, Array]
        , "dependenciesFile": path
        , "recursive": Boolean
    },
        shortHands = {
            "h": ["--help"],
            "d": ["--dependency"],
            "r": ["--recursive"]
        },

        description = {
            "workspacepath": " Path to the workspace where the source code exists",
            "outputpath": " Path where the compiled output exists",
            "branch": " Branch name",
            "build": " Build version",
            "commit": " commit ID, as provided by the SCM (e.g. SHA1 for git, revision for SVN, changeset for TFS). May be automatically detected by the SCM.",
            "appname": " Application Name",
            "technology": " The technology in which the project is written (more than one value may be specified)",
            "scm": " The Source Control Management used (currently only 'git' is supported)",
            "help": " Show this help page",
            "dependency": " Project dependencies. Pattern should be a semicolon-separated AppName@Branch@Build list. May be specified more than once.",
            "dependenciesFile": " A path to a json file that is in the following format: [{\"appName\":\"\",\"branch\":\"\",\"build\":\"\"},{...}]",
            "recursive": "Starts a recursive search for JARs & WARs in folder specified by the 'workspacepath' option"
        },

        defaults = {
            "help": "",
        };
    
    var buildArguments = nopt(knownOpts, shortHands, process.argv, 2);
    ["branch", "build", "appname"].forEach(function (key) {
        if (!buildArguments[key]) {
            console.error("Required argument '" + key + "' not specified");
            hasMissingArguments = true;
        }
    });
    
    var hadInvalidArguments = false;
    
    //workspacepath, technology, scm must all come together or not at all
    var count = 0;
    ["workspacepath", "technology", "scm"].forEach(function (key) {
        if (buildArguments[key]) {
            count++;
        }
    });
    
    if (count != 0 && count != 3) {
        console.error("The 'workspacepath', 'technology', 'scm' command-line arguments must all be specified together or not at all");
        hadInvalidArguments = true;
    }
    
    var parsedDependencies = [];
    if (buildArguments.dependency && !hadInvalidArguments) {
        try {
            parsedDependencies = parseDependencies(buildArguments.dependency);
        }
        catch (e) {
            console.error(e);
            hadInvalidArguments = true;
        }
    }
    
    if (buildArguments.dependenciesFile && !hadInvalidArguments) {
        try {
            var parsedJson = readJsonSync(buildArguments.dependenciesFile);
            if (!parsedJson || !Array.isArray(parsedJson)) {
                throw 'Invalid file structure';
            }
            
            parsedJson.forEach(function (depItem) {
                if (depItem.appName && depItem.build && depItem.branch) {
                    parsedDependencies.push({ appName: depItem.appName.toString(), branch: depItem.branch.toString(), build: depItem.build.toString() });
                } else {
                    throw 'Invalid dependency item: ' + JSON.stringify(depItem);
                }
            });

        } catch (e) {
            console.error('Error reading/parsing ' + buildArguments.dependenciesFile + ": " + e);
            hadInvalidArguments = true;
        }
    }
    
    if (hasMissingArguments || hadInvalidArguments || buildArguments.help) {
        try {
            var usage = noptUsage(knownOpts, shortHands, description, defaults);
            console.error(usage);
        } catch (e) {
            console.error(e);
        }
        callback(1);
    }
    
    var technologies = this.getTechnologies();
    var sourceConrolProviders = this.getSourceConrolProviders();
    var cfg = cfgLoader(buildArguments.config).then(function (cfg) {
        if (buildArguments.newBuildScanner) {
            var newDiffService = new NewDiffService(cfg, needle);
            var buildScanningProcess = new BuildScannerProcess(cfg, newDiffService, technologies, sourceConrolProviders, parsedDependencies);
            buildScanningProcess.run(buildArguments).then(function () {
                console.log('All done!');
                callback(0);
            });
        }
        else {
            var diffService = new DiffService(cfg, needle);
            var buildDiffProcess = new BuildDiffProcess(cfg, diffService, technologies, sourceConrolProviders, parsedDependencies);
            buildDiffProcess.run(buildArguments).then(function () {
                console.log('All done!');
                callback(0);
            });
        }
    });
}

module.exports.CIA = CIA;