﻿var uuid = require('uuid');
var methodKeyCreator = require('./method-key.js');
//var MappingFile = require('./mapping-file.js');
var statuses = require('../statuses.js');
const DEL = -1;
const INS = 1;

/** 
 * @description Expands multiple diff chunks to the line level
 * @returns {array} an array of {line,op} (op is either DEL or INS)
 **/
function expandDiffChunks(diffChunks) {
    var ops = [];
    var carry = 0;
    diffChunks.forEach(function (c) {
        var oldLine = c.oldStart + carry;
        for (var i = 0; i < c.changes.length; i++) {
            var l = c.changes[i];
            if (l.normal) {
                oldLine++;
            }
            else if (l.del) {
                ops.push({ line: oldLine, op: DEL });
            }
            else if (l.add) {
                ops.push({ line: oldLine, op: INS });
                oldLine++;
            }
            else throw 'Invalid diff line';
        }
        carry += (c.newLines - c.oldLines);

    });
    return ops;
}

function applyDiffChangesToMethods(oldMapping, diffFromOldToNew) {
    
    var expandedDiff = expandDiffChunks(diffFromOldToNew.chunks);
    var carry = 0;
    
    var methods = oldMapping.methods.filter(function (t) {return !!t.position });
    
    //oldMapping.methods.forEach(function (m) {        
    //    methods.push({ id: methodId, name: m.name, start: m.position/*, end: m.end*/ });
    //});
    
    methods.sort(function (m1, m2) {
        if (m1.position[0] < m2.position[0]) //m1 line is before m2 line
            return -1;
        if (m1.position[0] > m2.position[0]) //m1 line is after m2 line
            return 1;
        //m1, m2 are on the same line
        if (m1.position[1] < m2.position[1]) //m1 column is before m2 column
            return -1;
        if (m1.position[1] > m2.position[1]) //m1 column is after m2 column
            return 1;
        
        return 0; //same line, column
    });
    
    ///**
    // * @description counts the offset where the old lineNumber moved (positive if moved down, negative if moved up)
    // * INEFFICIENT, especially for large files. I wanted the algorithm to be correct first, then optimize it later as needed.
    // */
    //function accumulateOffsetBeforeLine(lineNumber) {
    //    var offset = 0;
    //    for (var i = 0; i < expandedDiff.length; i++) {
    //        if (expandedDiff[i].line > lineNumber) break;
    //        var d= expandedDiff[i].op == INS ? 1 : -1;
    //        offset += d;
    //        lineNumber += d;
    //    }
    //    return offset;
    //};
    
    /**
     * @description finds the position of the given (old, before the change) line number 
     * INEFFICIENT, especially for large files. I wanted the algorithm to be correct first, then optimize it later as needed.
     * Optimization tip - since the method is guaranteed to be called with increasing line numbers (methods are sorted that way), it's possible to memorize the last line and index in the expanded diff
     */
    function calcNewLineNumber(lineNumber) {
        for (var i = 0; i < expandedDiff.length; i++) {
            if (expandedDiff[i].line > lineNumber) break;
            var d = expandedDiff[i].op == INS ? 1 : -1;
            lineNumber += d;
        }
        return lineNumber;
    };
    
    /**
     * @description returns true iff there was a deletion at the given (old, before the change) line number 
     * INEFFICIENT, especially for large files. I wanted the algorithm to be correct first, then optimize it later as needed.
     * Optimization tip - since the method is guaranteed to be called with increasing line numbers (methods are sorted that way), it's possible to memorize the last line and index in the expanded diff
     */
    function hasDeletionsInLine(lineNumber) {
        for (var i = 0; i < expandedDiff.length; i++) {
            if (expandedDiff[i].line > lineNumber) break;
            if (expandedDiff[i].line == lineNumber && expandedDiff[i].op == DEL)
                return true;
            var d = expandedDiff[i].op == INS ? 1 : -1;
            lineNumber += d;
        }
        return false;
    }
    
    methods.forEach(function (method) {
        method.originalPosition = [method.position[0], method.position[1]];
        if (hasDeletionsInLine(method.position[0])) {
            method.position[0] = -1;
            method.position[1] = -1;
            return;
        }
        method.position[0] = calcNewLineNumber(method.position[0]);//+= accumulateOffsetBeforeLine(method.start[0]);
    });
}

function compareOldVsNew(oldMapping, newSignature) {
    newSignature.status = (oldMapping.hash == newSignature.hash) ? statuses.FILE_NOT_CHANGED : statuses.FILE_HASH_CHANGED;
    
    var oldLookup = {};
    
    oldMapping.methods.forEach(function (m) {
        var wholeKey = methodKeyCreator(m.position, m.name, m.isAnonymous);
        oldLookup[wholeKey] = m;
    });
    
    var newLookup = {};
    newSignature.methods.forEach(function (m) {
        var wholeKey = methodKeyCreator(m.position, m.name, m.isAnonymous);
        newLookup[wholeKey] = m;
    });
    
    Object.keys(oldLookup).forEach(function (methodKey) {
        var oldHash = oldLookup[methodKey].hash;
        var oldId = oldLookup[methodKey].id;
        
        var newMatch = newLookup[methodKey];
        if (newMatch) {
            newMatch.id = oldId;
            //compare hashes            
            var newHash = newMatch.hash;
            if (oldHash != newHash) {
                //method changed
                newLookup[methodKey].status = statuses.METHOD_CHANGED;
            }
            else {
                var oldPosition = (oldLookup[methodKey].originalPosition || oldLookup[methodKey].position).join(',');
                var newPosition = newLookup[methodKey].position.join(',');
                var positionChanged = oldPosition != newPosition;
                //method unchanged
                newLookup[methodKey].status = positionChanged ? statuses.METHOD_MOVED : statuses.METHOD_NOT_CHANGED;
            }
            
            delete newLookup[methodKey];
        } else {
            //method deleted
            //changes.removed.push(oldId);
            newSignature.methods.push({ id: oldId, status: statuses.METHOD_DELETED });
        }
    });
    
    Object.keys(newLookup).forEach(function (methodKey) {
        var newData = newLookup[methodKey];
        //method added - Generate a new ID
        var newId = uuid([methodKey, newData.hash, new Date().valueOf()].join('|'));
        newData.id = newId;
        newData.status = statuses.METHOD_ADDED;
    });
    
    return newSignature;
}

/**
 * @description Calculates diffs between two revisions of a single file (old/new)
 * @param {object} oldMapping old metada from server
 * @param {string} standard diff between the two revisions of the file
 * @param {string} newSignature signature of new file
 * @param {string} newRevision The new revision that will be stored in the mapping file
 * @return {object} changes in standard SeaLights format
 */
function fileDiff(oldMapping, diffFromOldToNew, newSignature, newRevision) {
    
    try {
        if (diffFromOldToNew) {
            //1. Run with two pointers over the old metadata and the diff to map each "old" method to its potential new location using the diff data only
            applyDiffChangesToMethods(oldMapping, diffFromOldToNew);
        }
        
        //2. Now, compare methods
        compareOldVsNew(oldMapping, newSignature);

    } catch (e) {
        console.log('FileDiff error: ' + e + "\n" + e.stack);
    }
    
    return newSignature;
}

module.exports = fileDiff;