﻿var esprima = require('esprima'),
    escodegen = require('escodegen'),
    traverse = require('ast-traverse'),
    astUtils = require('esprima-ast-utils'),
    md5 = require('md5'),
    SnippetProvider = require('./snippet-provider.js').SnippetProvider,
    getTransientMethodName = require('./transient-method-naming.js').getTransientMethodName,
    guessMethodName = require('./guess-method-name.js').guessMethodName;

function formatLoc(loc) {
    return formatLocToArray(loc).join(',');
}

function formatLocToArray(loc) {
    return [loc.line, loc.column];
}

function calculateFunctionNodeHash(node) {
    var clonedNode = astUtils.clone(node);
    //calculate body hash
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;
            if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                node.body = { type: "BlockStatement", body: [] };
                node.params = [];
                node.id = { name: '_' };
            }
        }
    });
    
    var generated = escodegen.generate(clonedNode);
    var hash = md5(generated);
    return hash;
}

/**
 * Creates a new Uint8Array based on two different ArrayBuffers
 * source: https://gist.github.com/72lions/4528834
 * @private
 * @param {ArrayBuffers} buffer1 The first buffer.
 * @param {ArrayBuffers} buffer2 The second buffer.
 * @return {ArrayBuffers} The new ArrayBuffer created out of the two.
 */
function removeHashbangPrefix(bytes) {
    function concatArrays(buffer1, buffer2) {
        var tmp = new Uint8Array(buffer1.byteLength + buffer2.byteLength);
        tmp.set(new Uint8Array(buffer1), 0);
        tmp.set(new Uint8Array(buffer2), buffer1.byteLength);
        return tmp;
    };

    var HASHBANG_BYTES = [35, 33];
    var LINE_COMMENT_BYTES = [47, 47];
    if (bytes.length > 2 && bytes[0] == HASHBANG_BYTES[0] && bytes[1] == HASHBANG_BYTES[1]) { //hashbang
        bytes = concatArrays(new Uint8Array(LINE_COMMENT_BYTES), bytes);//insert two slashes to make that line a comment
    }
    return bytes;
}

function getMethodName(node, parent){
    var mn = node.id && node.id.name;
    if (!mn) {
        mn = guessMethodName(node, parent) || "(Anonymous)";
    }
    //mn += " at " + [start.line, start.column].join(',');
    return mn;
}

function generateFileSignature(filename, bytes) {
    var result = {
        module: { hash: "" },
        methods: []
    };
    
    var snippetProvider = new SnippetProvider();
    snippetProvider.loadFile(filename, bytes);
    filename = filename.replace(/\\/g, '/'); //we want all transients to use unix / instead of windows \

    var ast;
    
    try {
        bytes = removeHashbangPrefix(bytes);

        //1. Parse the file
        ast = esprima.parse(bytes, {
            loc: true,
            range: true,
            comment: false //No need for those
        });

    } catch (parseException) {
        //TODO: Test
        return { error: "Parse error (1)" };
    }
    
    try {
        
        // print AST node types, pre-order (node first, then its children) 
        traverse(ast, {
            pre: function (node, parent, prop, idx) {
                if (node.type == "FunctionDeclaration" || node.type == "FunctionExpression") {
                    var hash = calculateFunctionNodeHash(node);
                    var start = formatLocToArray(node.loc.start);
                    var hasExplicitName = (node.id && node.id.name);
                    var name = getMethodName(node, parent);
                    var uniqueName = [name,filename,start.join(',')].join('@');
                    
                    result.methods.push({
                        name: name, //Method that can be presented to the user
                        isAnonymous: !hasExplicitName,
                        start: start,
                        //transientName: getTransientMethodName(filename, start.join(',')), //Name that's used for matching
                        hash: hash,
                        uniqueName: uniqueName
                        //snippet: snippetProvider.getSnippet(node.loc.start.line, node.loc.start.column)
                    })
                }
            }
        });
    }
    catch (traverseException1) {
        return { error: 'Parse error(2)' }
    }
    
    try {
        var bodyHash = calculateFunctionNodeHash(ast);
        result.module.hash = bodyHash;
    } catch (traverseException2) {
        return { error: "Parse error (3)" };
    }
    
    return result;
}

module.exports = generateFileSignature;