﻿/*
	ILogger interface:
		header(headerText:string): void;
		log(message:string): void;
		warn(message: string): void;
		error(message: string, e:any): void;
		attachFile(path:string, description: string): void;		
		createNestedLogger(name:string): ILogger;
		TODO: Add link(href,title): void;
*/

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * Null logger. Does not output anything.
 */
function NullLogger() {
}
NullLogger.prototype.header = function (headerText) { };
NullLogger.prototype.log = function (message) { };
NullLogger.prototype.warn = function (message) { };
NullLogger.prototype.error = function (message, e) { };
NullLogger.prototype.attachFile = function (path, description) { };
NullLogger.prototype.createNestedLogger = function (name) { return this; /* No point in a new instance */ };
module.exports.NullLogger = NullLogger;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * Console logger. Outputs to the console.
 * Nested loggers will have an indentation of a '\t' characted
 */
function ConsoleLogger(indentationLevel, name) {
    indentationLevel = indentationLevel || 0;
    var indentationStr = repeatString(indentationLevel, '\t');
    if (name)
        indentationStr += '[' + name + '] ';
    
    this.getIndentation = function () { return indentationLevel; }
    this.getIndentationStr = function () { return indentationStr; }
    this.getName = function () { return name; }
}
ConsoleLogger.prototype.header = function (headerText) {
    console.log('--------------------------------------------------------------------------------------------------');
    console.log(headerText);
    console.log('--------------------------------------------------------------------------------------------------');
};

ConsoleLogger.prototype.log = function (message) {
    console.log(this.getIndentationStr() + message);
};
ConsoleLogger.prototype.warn = function (message) {
    console.warn(this.getIndentationStr() + message);
};
ConsoleLogger.prototype.error = function (message, e) {
    if (e!==undefined)
        console.error(this.getIndentationStr() + message, e);
    else
        console.error(this.getIndentationStr() + message);
};
ConsoleLogger.prototype.attachFile = function (path, description) {
    console.log(this.getIndentationStr() + description + ": Attachment in " + path);
};
ConsoleLogger.prototype.createNestedLogger = function (name) {
    return new ConsoleLogger(this.getIndentation() + 1, name);
};

module.exports.ConsoleLogger = ConsoleLogger;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

/**
 * Sends log events to multiple loggers
 */
function MultipleLogger() {
    var loggers = [];
    this.addLogger = function (logger) { loggers.push(logger); }
    this.eachLogger = function (fn) {
        loggers.forEach(fn);
    }
}
MultipleLogger.prototype.header = function (headerText) {
    this.eachLogger(function (l) { l.header(headerText); });
};
MultipleLogger.prototype.log = function (message) {
    this.eachLogger(function (l) { l.log(message); });
};
MultipleLogger.prototype.warn = function (message) {
    this.eachLogger(function (l) { l.warn(message); });
};
MultipleLogger.prototype.error = function (message, e) {
    this.eachLogger(function (l) { l.error(message, e); });
};
MultipleLogger.prototype.attachFile = function (path, description) {
    this.eachLogger(function (l) { l.attachFile(path, description); });
};

MultipleLogger.prototype.createNestedLogger = function (name) {
    var sl = new MultipleLogger();
    this.eachLogger(function (l) {
        var nl = l.createNestedLogger(name);
        sl.addLogger(nl);
    });
    return sl;
};
module.exports.MultipleLogger = MultipleLogger;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

function MemoryLogger() {
    var buffer = [];
    
    this.createNestedLogger = function (name) {
        var ml = new MemoryLogger();
        buffer.push({ type: 'logger', instance: ml, name: name });
        return ml;
    }
    
    this.header = function (headerText) {
        buffer.push({ type: 'header', message: headerText });
    }
    
    this.log = function (message) {
        buffer.push({ type: 'log', message: message });
    };
    
    this.warn = function (message) {
        buffer.push({ type: 'warn', message: message });
    };
    
    this.error = function (message, e) {
        buffer.push({ type: 'error', message: message, exception: e });
    };
    
    this.attachFile = function (path, description) {
        buffer.push({ type: 'attachment', path: path, description: description });
    };
    
    function internalToString(prefix, name, formatter) {
        prefix = prefix || '';
        var prefixWithName = prefix;
        if (name) {
            prefixWithName += '[' + name + '] ';
        }
        var sb = [];
        buffer.forEach(function (item) {
            switch (item.type) {
                case 'header':
                    sb.push(formatter(prefixWithName, 'header', item.message));
                    break;

                case 'log':
                    sb.push(formatter(prefixWithName, 'log', item.message));
                    break;

                case 'warn':
                    sb.push(formatter(prefixWithName, 'warn', item.message));
                    break;

                case 'error':
                    sb.push(formatter(prefixWithName, 'error', item.message));
                    if (item.exception)
                        sb.push(formatter(prefixWithName, 'error', item.exception.toString()));
                    break;

                case 'attachment':
                    sb.push(formatter(prefixWithName, 'file', item.description + " at " + item.path));
                    break;

                case 'logger':
                    var delim = formatter.prefix || '\t';
                    sb.push(item.instance.toString(prefix + delim, item.name, formatter));
                    break;

                default:
                    sb.push(formatter(prefixWithName, 'unkwn', JSON.stringify(item, null, '\t')));
                    break;
            }
        });
        return sb.join('\n');
    }
    
    this.formatAs = function (formatter) {
        return this.toString(null, null, formatter);
    }
    
    this.toString = function (prefix, name, formatter) {
        formatter = formatter || this.StringFormatter;
        if (!isFunction(formatter)) {
            switch (formatter.toString().toLowerCase()) {
                case 'markdown':
                    formatter = this.MarkdownFormatter;
                    break;
                case '':
                default:
                    formatter = this.StringFormatter;
                    break;
            }
        }
        return internalToString(prefix, name, formatter);
    }
    
    this.StringFormatter = function (pfx, type, msg) {
        if (!msg) return '';
        switch (type) {
            case 'header':
                return [
                    pfx + '--------------------------------------------------------------------------------------------------',
                    pfx + msg,
                    pfx + '--------------------------------------------------------------------------------------------------',
                ].join('\n');
            default:
                return pfx + '[' + type + '] ' + msg;
        }
    }
    this.StringFormatter.delim = '\t';
    
    this.MarkdownFormatter = function (pfx, type, msg) {
        if (!msg) return '';
        switch (type) {
            case 'header':
                {
                    var indentationLevel = countTabPrefixLength(pfx);
                    return repeatString(indentationLevel, '#') + ' ' + msg; //prefix with ## for a <h1> or <h2> tag (or <h3> or <h4>, depending on the prefix length. Prefix should be tabs)
                }
            case 'log':
                {
                    if (msg.indexOf('\n')) {
                        msg = "<pre>" + msg + "</pre>";
                    }
                    return pfx + msg;
                }
            case 'warn':
                {
                    if (msg.indexOf('\n')) {
                        msg = "<pre>" + msg + "</pre>";
                    }
                    return pfx + ':weary:' + msg; //prefix with the :weary: emoji
                }
            case 'error':
                {
                    if (msg.indexOf('\n')) {
                        msg = "<pre>" + msg + "</pre>";
                    }
                    return pfx + ':boom:**' + msg + '**'; //Bold with the :boom: emoji
                }
            case 'attachment':
                return pfx + ':package:' + msg; //prefix with the :package: emoji
            default:
                return pfx + '[' + type + '] ' + msg;
        }
    }
    this.MarkdownFormatter.delim = '\t';
}
module.exports.MemoryLogger = MemoryLogger;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function countTabPrefixLength(str) {
    var i = 0;
    while (i < str.length) {
        if (str[i] == '\t') {
            i++;
        }
        else {
            break;
        }
    }
    return i;
}

function repeatString(count, str) {
    return new Array(count + 1).join(str); //works well when indentationLevel==0
}
function isFunction(functionToCheck) {
    var getType = {};
    return functionToCheck && getType.toString.call(functionToCheck) === '[object Function]';
}

function isArray(someVar) {
    return (Object.prototype.toString.call(someVar) === '[object Array]');
}
function isObject(someVar) {
    return (Object.prototype.toString.call(someVar) === '[object Object]');
}
function isLogger(someVar) {
    return someVar && someVar.log && someVar.warn && someVar.error && someVar.attachFile && true;
}

function loggerFactory(type) {
    if (arguments.length > 1) {
        //passed multiple arguments. They may be either objects or strings
        var args = Array.prototype.slice.call(arguments);
        var ml = new MultipleLogger();
        args.forEach(function (loggerDefOrInstance) {
            loggerDefOrInstance = loggerFactory(loggerDefOrInstance);
            ml.addLogger(loggerDefOrInstance);
        });
        return ml;
    }
    if (type && isArray(type)) {
        //passed 1 argument with an array. Contents may be either objects or strings
        var args = type;
        var ml = new MultipleLogger();
        args.forEach(function (loggerDefOrInstance) {
            loggerDefOrInstance = loggerFactory(loggerDefOrInstance);
            ml.addLogger(loggerDefOrInstance);
        });
        return ml;
    }
    if (isLogger(type))
        return type; //should already be a logger
    
    type = type || '';
    switch (type.toLowerCase()) {
        case '':
        case 'null':
            return new NullLogger();
        case 'memory':
            return new MemoryLogger();
        case 'console':
            return new ConsoleLogger();
    }
}

module.exports.createLogger = loggerFactory;